/*****************************************************
 * File    : 	include/linux/ktau/ktau_hash.h
 * Version : 	"$Id: ktau_hash.h,v 1.3 2005/09/30 15:51:48 suravee Exp $;
 ****************************************************/

#ifndef _KTAU_HASH_H_
#define _KTAU_HASH_H_

#define KTAU_HASH_SIZE		1024
#define KTAU_NESTED_LEVEL	20	

#ifdef CONFIG_SMP
#include <linux/spinlock.h>	/* For spinlock */
#endif /* CONFIG_SMP */

#include "ktau_datatype.h"

/* Locking and disabling  mechanism in KTAU */
#define ktau_bh_enable()        if(!in_softirq())\
                                        local_bh_enable();\
                                else\
                                        __local_bh_enable()


#ifdef CONFIG_KTAU_IRQ
#define ktau_spin_lock(l,f)     spin_lock_irqsave(l,f)
#define ktau_spin_unlock(l,f)   spin_unlock_irqrestore(l,f)
#else /*CONFIG_KTAU_IRQ*/
#define ktau_spin_lock(l,f)     local_bh_disable();\
                                spin_lock(l)
#define ktau_spin_unlock(l,f)   spin_unlock(l);\
                                ktau_bh_enable()
#endif /*CONFIG_KTAU_IRQ*/

/*
 * this struct is used in the task_struct
 */
typedef struct _ktau_prof{

	spinlock_t 		hash_lock;
	unsigned long 		prev_start;
	unsigned long 		prev_stop;
	unsigned int 		ent_count;
	unsigned int 		last_level;
	unsigned int 		cur_level;
	unsigned long long 	accum[KTAU_NESTED_LEVEL];
	h_ent* 			ktau_hash[KTAU_HASH_SIZE];
	ktau_state** 		ppstate;
#ifdef CONFIG_KTAU_PERF_COUNTER
	unsigned int 		accum_pcounter[KTAU_MAX_PERF_COUNTER][KTAU_NESTED_LEVEL];
#endif /*CONFIG_KTAU_PERF_COUNTER*/
	
#ifdef CONFIG_KTAU_TRACE
	ktau_trace* 		trbuf;
	ktau_trace* 		cur_trace ;
	ktau_trace* 		head_trace ;
	ktau_trace*		last_dump;
	ktau_trace		first_lost;
	ktau_trace		last_lost;
#endif /*CONFIG_KTAU_TRACE*/

} ktau_prof; 

/*----------------------- PROFILE API -----------------------------*/
/*
 * IN APIs : These will be used in process creation
 * 	     and destruction.
 */
extern void create_task_profile(struct task_struct *tsk);
extern void remove_task_profile(struct task_struct *tsk);

/* GLOBAL INDEX */
extern void reset_ktau_index(void);
void ktau_global_semaphore_init(void);

/*----------------------- MERGE API -----------------------------*/
/*
 * This function updates the uspace-state used 
 * for merging kernel & user profiles.
 *
 * Note: Does not perform ANY LOCKING. Hence any 
 * required locking must be performed externally.
 */
extern int update_userstate(unsigned long long incl, ktau_state** ppstate); 

/*----------------------- TRACE API -----------------------------*/
extern int trace_buf_init(struct task_struct *tsk);
extern int trace_buf_destroy(struct task_struct *tsk);
extern int trace_buf_write(unsigned long long tsc 
			 , unsigned long addr 
			 , unsigned long type);

#endif /*_KTAU_HASH_H_*/
