/*************************************************************************
 * File 	: kernel/ktau/ktau_inst.c
 * Version	: $Id: ktau_inst.c,v 1.3 2005/08/29 23:16:18 suravee Exp $
 *
 ************************************************************************/

/* to turn on debug messages */

#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/slab.h>			/* Kmalloc */	
#include <linux/sched.h>		/* task_struct def */
#include <linux/interrupt.h>
//#include <linux/hardirq.h>

#include <asm/atomic.h>			/* atomic */
#include <asm/current.h>		/* get_current */
#include <asm/system.h>			/* irqs_disabled */

#include <linux/spinlock.h>        	/* spinlock */

#define TAU_NAME "ktau_inst"
#define PFX TAU_NAME
//#define TAU_DEBUG 		1

#include <linux/ktau/ktau_print.h>
#include <linux/ktau/ktau_inst.h>
#include <linux/ktau/ktau_datatype.h>
#include <linux/ktau/ktau_hash.h>
#include <linux/ktau/ktau_perf.h>

#define KTAU_START_INDEX	399	/* Dynamic index start at 400 */

/*--------------------- GLOBAL INDEX  ---------------------*/
/*
 * This function is from linux-2.6 kernel since it does not
 * exist in linux-2.4.
 */
#ifdef CONFIG_SMP
#define LOCK "lock ; "
#else
#define LOCK ""
#endif


/*
 * This macro and fuction does not exist in 2.4.x for 
 * X86 architecture, but does exist for PPC. 
 */
#ifndef CONFIG_PPC32
#define atomic_inc_return(v)	(atomic_add_return(1,v))
static __inline__ int atomic_add_return(int i, atomic_t *v)
{
        int __i;
#ifdef CONFIG_M386
        if(unlikely(boot_cpu_data.x86==3))
                goto no_xadd;
#endif

	/* Modern 486+ processor */
        __i = i;
        __asm__ __volatile__(
                LOCK "xaddl %0, %1;"
                :"=r"(i)
                :"m"(v->counter), "r"(i));
        return i + __i;

#ifdef CONFIG_M386
no_xadd: /* Legacy 386 processor */
        local_irq_disable();
        __i = atomic_read(v);
        atomic_set(v, i + __i);
        local_irq_enable();
        return i + __i;
#endif /*CONFIG_M386*/
}
#endif /*CONFIG_PPC32*/


/*
 * Global function index
 */
atomic_t ktau_global_index = ATOMIC_INIT(KTAU_START_INDEX);

/* 
 * This function is used to increment the global_ktau_index.
 */
unsigned int get_ktau_index(void){
	return atomic_inc_return(&ktau_global_index);	
}

/*
 * Increment global function index by 'incr'
 * then Return the global function index for hash table
 * incremented by 1 from original value (i.e same return
 * value as from get_kau_index).
 */
unsigned int incr_ktau_index(int incr) {
        return atomic_add_return(incr, &ktau_global_index);
}

void reset_ktau_index(void){
	atomic_set(&ktau_global_index,KTAU_START_INDEX);
}
/*--------------------------- TIMER ---------------------------------*/

#ifdef CONFIG_X86   

/*
 * The low-level Pentium-class timer    
 */
static __inline__ unsigned long long int ktau_rdtsc(void)
{
     unsigned long long int x;
     __asm__ volatile (".byte 0x0f, 0x31" : "=A" (x));
     return x;
}

#else /*CONFIG_X86*/
#ifdef CONFIG_PPC32
/*
 * The low-level PowerPC timer (TimeBase)   
 */
//static __inline__ unsigned long long ktau_rdtsc(void) {
unsigned long long ktau_rdtsc(void) {

        unsigned long long time = 0;
        unsigned long lword = 0, hword = 0;

        asm volatile (
                "rdtscppc: \n\t"
                "mfspr  %%r27, 269 \n\t"
                "mfspr  %%r28, 268 \n\t"
                "mfspr  %%r29, 269 \n\t"
                "cmpw   %%r29, %%r27 \n\t"
                "bne    rdtscppc \n\t"
                "ori    %0,%%r28,0 \n\t"
                "ori    %1,%%r29,0 \n\t"
                : "=r" (lword), "=r" (hword)
                :
                : "r27", "r28", "r29"
        );

        time = hword;

        return ( (time << 32) | lword );
}
#endif /*CONFIG_PPC32*/
#endif /*CONFIG_X86*/



/*--------------------- INSTRUMENTATION ---------------------*/

#ifdef CONFIG_KTAU_PERF_COUNTER
/* 
 * ktau_perf_counter diff variable 
 */
unsigned int diff_pcounter[KTAU_MAX_PERF_COUNTER] = {0};
unsigned int ktau_pcounter_mask = 0;
unsigned int ktau_num_pcounter = 0;

/*
 * To reset ktau_perf_counter
 */
void reset_pcounters(h_ent *cur_ent){
	int i=0;
#ifdef CONFIG_KTAU_PERF0
	i++;
	cur_ent->pcounter[i].count	= 0;	
	cur_ent->pcounter[i].start	= 0;		
	cur_ent->pcounter[i].incl	= 0;		
	cur_ent->pcounter[i].excl	= 0;	
#endif /*CONFIG_KTAU_PERF0*/
#ifdef CONFIG_KTAU_PERF1
	i++;
	cur_ent->pcounter[i].count	= 0;		
	cur_ent->pcounter[i].start	= 0;		
	cur_ent->pcounter[i].incl	= 0;		
	cur_ent->pcounter[i].excl	= 0;		
#endif /*CONFIG_KTAU_PERF1*/
#ifdef CONFIG_KTAU_PERF2
	i++;
	cur_ent->pcounter[i].count	= 0;		
	cur_ent->pcounter[i].start	= 0;		
	cur_ent->pcounter[i].incl	= 0;		
	cur_ent->pcounter[i].excl	= 0;		
#endif /*CONFIG_KTAU_PERF2*/
#ifdef CONFIG_KTAU_PERF3
	i++;
	cur_ent->pcounter[i].count	= 0;		
	cur_ent->pcounter[i].start	= 0;		
	cur_ent->pcounter[i].incl	= 0;		
	cur_ent->pcounter[i].excl	= 0;		
#endif /*CONFIG_KTAU_PERF3*/
#ifdef CONFIG_KTAU_PERF4
	i++;
	cur_ent->pcounter[i].count	= 0;		
	cur_ent->pcounter[i].start	= 0;		
	cur_ent->pcounter[i].incl	= 0;		
	cur_ent->pcounter[i].excl	= 0;		
#endif /*CONFIG_KTAU_PERF4*/

	/*...... */
}

/*
 * To start ktau_perf_counter
 */
void start_pcounters(h_ent *cur_ent){
	int i=0;
	for(i=0 ; i < ktau_num_pcounter; i++){
		cur_ent->pcounter[i].start	= ktau_perf_func[i]();
	}
}

/*
 * To stop ktau_perf_counter
 */
void stop_pcounters(h_ent *cur_ent){
	int i = 0;
	for(i=0; i<ktau_num_pcounter; i++){
		cur_ent->pcounter[i].count++;
		diff_pcounter[i] 		= ktau_perf_func[i]();
		diff_pcounter[i] 		= diff_pcounter[i] - cur_ent->pcounter[i].start;
		cur_ent->pcounter[i].incl 	+= diff_pcounter[i];
		cur_ent->pcounter[i].start 	= 0;
	}
}


void update_pcounter_excl_1(h_ent *cur_ent){
	unsigned int i = 0;
	for(i=0;i<ktau_num_pcounter;i++){
		cur_ent->pcounter[i].excl += diff_pcounter[i];
	}
}

void update_pcounter_excl_2(h_ent *cur_ent){
	unsigned int i = 0;
	for(i=0;i<ktau_num_pcounter;i++){
		cur_ent->pcounter[i].excl += 
			diff_pcounter[i] - 
			current->ktau->accum_pcounter[i][current->ktau->last_level];
		current->ktau->accum_pcounter[i][current->ktau->last_level] = 0;
		current->ktau->accum_pcounter[i][current->ktau->cur_level] = 0;
	}
}

void update_pcounter_excl_3(h_ent *cur_ent){
	unsigned int i = 0;
	for(i=0;i<ktau_num_pcounter;i++){
		cur_ent->pcounter[i].excl += diff_pcounter[i];
		current->ktau->accum_pcounter[i][current->ktau->cur_level] = diff_pcounter[i];
	}
	
}
void update_pcounter_excl_4(h_ent *cur_ent){
	unsigned int i = 0;
	for(i=0;i<ktau_num_pcounter;i++){
		cur_ent->pcounter[i].excl += 
			diff_pcounter[i]- 
			current->ktau->accum_pcounter[i][current->ktau->last_level];
		current->ktau->accum_pcounter[i][current->ktau->last_level] = 0;
		current->ktau->accum_pcounter[i][current->ktau->cur_level] += diff_pcounter[i];
	}	
}
#endif /*CONFIG_KTAU_PERF_COUNTER*/

/*******************************************************************************************/
/*
 * This fuction is the start profile
 */
void ktau_start_prof(unsigned int index,unsigned int addr){
	unsigned long flags = 0;
	current->ktau->prev_start = addr;
	h_ent *cur_ent = (h_ent*)(current->ktau->ktau_hash[index]);
	
	ktau_spin_lock(&current->ktau->hash_lock,flags);
	
	/********************************************************/
	/* 
	 * Incl / Excl time logic 
	 */	
#ifdef CONFIG_KTAU_MERGE
	if(current->ktau->ppstate) info("STARTING: addr=%x , cur_level= %u , last_level= %u",
				addr,current->ktau->cur_level, current->ktau->last_level);
#endif /*CONFIG_KTAU_MERGE*/

	if(current->ktau->last_level == 0 && 
			current->ktau->cur_level == 0){
		/* Normal Case */
		current->ktau->cur_level = 1;
#ifdef CONFIG_KTAU_MERGE
		if(current->ktau->ppstate)info("---- Normal case");
#endif /*CONFIG_KTAU_MERGE*/
	}else if(current->ktau->last_level < current->ktau->cur_level){
		/* Enter Nested Case */
		current->ktau->last_level++;
		current->ktau->cur_level++;
#ifdef CONFIG_KTAU_MERGE
		if(current->ktau->ppstate)info("---- Enter nested case");
#endif /*CONFIG_KTAU_MERGE*/
	}else if(current->ktau->last_level > current->ktau->cur_level){
		/* Re-enter Nested Case */
		current->ktau->last_level--;
		current->ktau->cur_level++;
#ifdef CONFIG_KTAU_MERGE
		if(current->ktau->ppstate)info("---- Re-enter nested case");
#endif /*CONFIG_KTAU_MERGE*/
	}else{
		printk("ktau_start_prof: Invalid instrumentation level");	
	}
	/********************************************************/
	/*
	 * Update hash table
	 */
	if(!(cur_ent)){
		current->ktau->ktau_hash[index] = kmalloc(sizeof(h_ent),GFP_ATOMIC);
		current->ktau->ent_count++;
		cur_ent = current->ktau->ktau_hash[index];
		cur_ent->type 			= KTAU_TIMER;
		cur_ent->data.timer.count 	= 0;
		cur_ent->data.timer.incl 	= 0;
		cur_ent->data.timer.excl 	= 0;
		cur_ent->addr 	 		= addr;
#ifdef CONFIG_KTAU_PERF_COUNTER
		reset_pcounters(cur_ent);
#endif /*CONFIG_KTAU_PERF_COUNTER*/
	}
	cur_ent->data.timer.start = ktau_rdtsc();
#ifdef CONFIG_KTAU_PERF_COUNTER
	start_pcounters(cur_ent);
#endif /*CONFIG_KTAU_PERF_COUNTER*/

#ifdef CONFIG_KTAU_TRACE
	/*
	 * Update trace buffer
	 */
	trace_buf_write(cur_ent->data.timer.start, addr, KTAU_TRACE_START);
#endif /*CONFIG_KTAU_TRACE*/

	/********************************************************/
	ktau_spin_unlock(&current->ktau->hash_lock,flags);
}

/*
 * This fuction is the stop profile
 */
void ktau_stop_prof(unsigned int index, unsigned int addr){
	unsigned long flags = 0;
	unsigned long long diff_time = 0;
	current->ktau->prev_stop = addr;

	ktau_spin_lock(&current->ktau->hash_lock,flags);

	/********************************************************/
	h_ent *cur_ent = (h_ent*)(current->ktau->ktau_hash[index]);
	if((cur_ent)){
		cur_ent->data.timer.count++;
		diff_time = ktau_rdtsc();
#ifdef CONFIG_KTAU_PERF_COUNTER
		stop_pcounters(cur_ent);
#endif /*CONFIG_KTAU_PERF_COUNTER*/

#ifdef CONFIG_KTAU_TRACE
		/*
		 * Update trace buffer
		 */
		trace_buf_write(diff_time, addr, KTAU_TRACE_STOP);
#endif /*CONFIG_KTAU_TRACE*/

		/*
		 * Update hash table
		 */
		diff_time = diff_time - cur_ent->data.timer.start;
		cur_ent->data.timer.incl += diff_time;
		cur_ent->data.timer.start = 0;

#ifdef CONFIG_KTAU_MERGE
		if(current->ktau->ppstate) info("STOPPING: addr=%x , cur_level= %u , last_level= %u",
					addr,current->ktau->cur_level, current->ktau->last_level);
#endif /*CONFIG_KTAU_MERGE*/

		/********************************************************/
		/* 
		 * Incl / Excl time logic 
		 */	
		if(current->ktau->last_level == 0 && 
				current->ktau->cur_level == 1){
			/*Normal case */
			current->ktau->last_level 	= 0;
			current->ktau->cur_level 	= 0;
			cur_ent->data.timer.excl  	+= diff_time;
#ifdef CONFIG_KTAU_PERF_COUNTER
			update_pcounter_excl_1(cur_ent);
#endif /*CONFIG_KTAU_PERF_COUNTER*/

#ifdef CONFIG_KTAU_MERGE
			if(current->ktau->ppstate) info("---- Normal Case");
#endif /*CONFIG_KTAU_MERGE*/
		}else if(current->ktau->last_level == 2 && 
				current->ktau->cur_level == 1){
			/* Leaving to level 0 */
			cur_ent->data.timer.excl 	+= 
				diff_time - current->ktau->accum[current->ktau->last_level];
			current->ktau->accum[current->ktau->last_level] = 0;
			current->ktau->accum[current->ktau->cur_level]  = 0;
#ifdef CONFIG_KTAU_PERF_COUNTER
			update_pcounter_excl_2(cur_ent);
#endif /*CONFIG_KTAU_PERF_COUNTER*/

			current->ktau->last_level 	= 0;
			current->ktau->cur_level 	= 0;

#ifdef CONFIG_KTAU_MERGE
			if(current->ktau->ppstate) info("---- Leaving to level 0");
#endif /*CONFIG_KTAU_MERGE*/
		}else if(current->ktau->last_level < current->ktau->cur_level){
			/* Leaving from inner most nested */
			cur_ent->data.timer.excl 	+= diff_time;
			current->ktau->accum[current->ktau->cur_level]  = diff_time;
#ifdef CONFIG_KTAU_PERF_COUNTER
			update_pcounter_excl_3(cur_ent);
#endif /*CONFIG_KTAU_PERF_COUNTER*/

			current->ktau->last_level++;
			current->ktau->cur_level--;
	
#ifdef CONFIG_KTAU_MERGE
			if(current->ktau->ppstate) info("---- Leaving from inner most nested");
#endif /*CONFIG_KTAU_MERGE*/
		}else if(current->ktau->last_level > current->ktau->cur_level){
			/* Leaving Nested */
			cur_ent->data.timer.excl 	+= 
				diff_time - current->ktau->accum[current->ktau->last_level];
			current->ktau->accum[current->ktau->last_level] = 0;
			current->ktau->accum[current->ktau->cur_level]  += diff_time;
#ifdef CONFIG_KTAU_PERF_COUNTER
			update_pcounter_excl_4(cur_ent);
#endif /*CONFIG_KTAU_PERF_COUNTER*/

			current->ktau->last_level--;
			current->ktau->cur_level--;

#ifdef CONFIG_KTAU_MERGE
			if(current->ktau->ppstate)info("---- Leaving Nested");
#endif /*CONFIG_KTAU_MERGE*/
		}else{
			printk("ktau_stop_prof: Invalid instrumentation level");	
		}
		/********************************************************/
	}
	/********************************************************/

	ktau_spin_unlock(&current->ktau->hash_lock,flags);

#ifdef CONFIG_KTAU_MERGE
	/* Update ktau_state once get to the outer most level */
	if(current->ktau->cur_level == 0 && current->ktau->last_level == 0)
		if(current->ktau->ppstate) update_userstate(diff_time, current->ktau->ppstate);
#endif /*CONFIG_KTAU_MERGE*/
}

/*
 * This fuction is the event profile
 */
void ktau_event_prof(unsigned int index, unsigned int addr){
	unsigned long flags = 0;
	h_ent *cur_ent = (h_ent*)(current->ktau->ktau_hash[index]);

	ktau_spin_lock(&current->ktau->hash_lock,flags);

	if(!(cur_ent)){
		current->ktau->ktau_hash[index] = kmalloc(sizeof(h_ent),GFP_ATOMIC);
		current->ktau->ent_count++;
		cur_ent = current->ktau->ktau_hash[index];
		cur_ent->type = KTAU_EVENT;
		cur_ent->data.timer.count = 0;
		cur_ent->data.timer.incl = 0;
		cur_ent->data.timer.start = 0;
		//cur_ent->data.timer.addr  = addr;
	}
	cur_ent->data.timer.count++;

	ktau_spin_unlock(&current->ktau->hash_lock,flags);
}

