/*****************************************************
 * File    : 	include/linux/ktau/ktau_hash.h
 * Version : 	"$Id: ktau_hash.h,v 1.1 2005/06/29 16:01:19 suravee Exp $;
 ****************************************************/

#ifndef _KTAU_HASH_H_
#define _KTAU_HASH_H_

#define KTAU_HASH_SIZE		1024
#define KTAU_NESTED_LEVEL	20	

#ifdef CONFIG_SMP
#include <linux/spinlock.h>	/* For spinlock */
#endif /* CONFIG_SMP */

#include "ktau_datatype.h"

/*
 * this struct is used in the task_struct
 */
typedef struct _ktau_prof{
#ifdef CONFIG_SMP
	spinlock_t hash_lock;
#endif /* CONFIG_SMP */
	unsigned long long tsc_rd;
	unsigned long long tsc_wr;
	unsigned int ent_count;
	unsigned int last_level;
	unsigned int cur_level;
	unsigned long long accum[KTAU_NESTED_LEVEL];
	h_ent* ktau_hash[KTAU_HASH_SIZE];

	ktau_state** ppstate;

} ktau_prof; 

/*
 * IN APIs : These will be used in process creation
 * 	     and destruction.
 */
extern void create_task_profile(struct task_struct *tsk);
extern void remove_task_profile(struct task_struct *tsk);

/* GLOBAL INDEX */
extern void reset_ktau_index(void);
void ktau_global_semaphore_init(void);

/*
 * This function updates the uspace-state used 
 * for merging kernel & user profiles.
 *
 * Note: Does not perform ANY LOCKING. Hence any 
 * required locking must be performed externally.
 */
extern int update_userstate(unsigned long long incl, ktau_state** ppstate); 

#endif /*_KTAU_HASH_H_*/
