/***********************************************
 * File		: include/linux/ktau/ktau_proc_data.h
 * Version	: $Id: ktau_proc_data.h,v 1.1 2005/06/29 16:01:19 suravee Exp $
 ***********************************************/

#ifndef _KTAU_PROC_DATA_H
#define _KTAU_PROC_DATA_H

#include <linux/ktau/ktau_datatype.h>

/* User-Space Application Data-Format */

/* IMPORTANT: How do we make sure NO    *
 * Padding is added to below structures *
 * ?
 */

/* Flags */
////////////////////////

/* Input Flags - that you can pass to ktau */
	/* none currently */

/* Output Flags - that ktau returns */
#define KTAU_SUCCESS		(1 << 0) /* all is well */
#define KTAU_TOLERANCE_PASS	(1 << 1) /* tolerance was fine, so no data was passed back */


/* Data types */
////////////////////////

typedef struct _ktau_tolerance {
	unsigned long tol_in;		/* in-arg: tolerance for duplicate reads */
	unsigned long time;		/* out-arg: last read when */
} ktau_tolerance_t;


typedef struct _ktau_size {
	unsigned int flags;
	ktau_tolerance_t tol;
	unsigned long size;		/* out-arg: size of profile */	
} ktau_size_t;


typedef struct _ktau_read {
	unsigned int flags;
	ktau_tolerance_t tol;
	unsigned long size;		/* in-arg: size of data buffer */	
	char* data;			/* out-arg: profile data */
} ktau_read_t;


/* Currently the size & buffer fields of ktau_write_t *
 * are NOT USED. write will ONLY purge (if tolerance  *
 * is correct ). */

typedef struct _ktau_write {
	unsigned int flags;
	ktau_tolerance_t tol;
	unsigned long size;		/* in-arg: size of buffer */	
	char* buffer;			/* in-arg: data to write */
} ktau_write_t;

typedef struct _ktau_merge {
	unsigned int flags;
	ktau_tolerance_t tol;
	ktau_state** ppstate;		/* in-arg: user-space state */	
} ktau_merge_t;

typedef union _ktau_proc_data {
	ktau_size_t size;
	ktau_read_t read;
	ktau_write_t write;
	ktau_merge_t merge;
} ktau_proc_data_t;

/* types -- just use the ioctl command types */
typedef struct _ktau_package {
	unsigned int type;
	ktau_proc_data_t op;
	pid_t pid; /*if pid<0 then Self */
} ktau_package_t;



#endif  /*_KTAU_PROC_DATA_H */



