/*************************************************************************
 * File 	: kernel/ktau/ktau_inst.c
 * Version	: $Id: ktau_inst.c,v 1.1 2005/06/29 16:03:03 suravee Exp $
 *
 ************************************************************************/

/* to turn on debug messages */

#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/slab.h>			/* Kmalloc */	
#include <linux/sched.h>		/* task_struct def */
#include <linux/interrupt.h>
#include <linux/hardirq.h>

#include <asm/atomic.h>			/* atomic */
#include <asm/current.h>		/* get_current */
#include <asm/system.h>			/* irqs_disabled */

#include <linux/spinlock.h>        	/* spinlock */

#define TAU_NAME "ktau_inst"
#define PFX TAU_NAME
//#define TAU_DEBUG 		1

#include <linux/ktau/ktau_print.h>
#include <linux/ktau/ktau_inst.h>
#include <linux/ktau/ktau_datatype.h>
#include <linux/ktau/ktau_hash.h>

#define KTAU_START_INDEX	399	/* Dynamic index start at 400 */

#define ktau_bh_enable()	if(!in_softirq())\
					local_bh_enable();\
				else\
					__local_bh_enable()

#ifdef CONFIG_KTAU_IRQ
#define ktau_irq_disable_save_flags_lock(f,l)\
				spin_lock_irqsave(l,f)
#define ktau_irq_enable_restore_flags_unlock(f,l)\
				spin_unlock_irqrestore(l,f)
#endif /*CONFIG_KTAU_IRQ*/

/*--------------------- GLOBAL INDEX  ---------------------*/
/* 
 * This function is used to increment the global_ktau_index.
 */
atomic_t ktau_global_index = ATOMIC_INIT(KTAU_START_INDEX);

unsigned int get_ktau_index(void){
	return atomic_inc_return(&ktau_global_index);	
}

void reset_ktau_index(void){
	atomic_set(&ktau_global_index,KTAU_START_INDEX);
}
/*--------------------------- TIMER ---------------------------------*/
/*
 * The low-level Pentium-class timer    
 */
static __inline__ unsigned long long int ktau_rdtsc(void)
{
     unsigned long long int x;
     __asm__ volatile (".byte 0x0f, 0x31" : "=A" (x));
     return x;
}

/*--------------------- INSTRUMENTATION ---------------------*/
/*
 * This fuction is the start profile
 */
void ktau_start_prof(unsigned int index,unsigned int addr){
	unsigned long flags = 0;
	h_ent *cur_ent = (h_ent*)(current->ktau->ktau_hash[index]);

#ifdef CONFIG_KTAU_IRQ
	ktau_irq_disable_save_flags_lock(flags, &current->ktau->hash_lock);
#else /* CONFIG_KTAU_IRQ */
	local_bh_disable();
	spin_lock(&current->ktau->hash_lock);
#endif /* CONFIG_KTAU_IRQ */

	/* Incl / Excl time logic */	
#ifdef CONFIG_KTAU_MERGE
	if(current->ktau->ppstate) info("STARTING: addr=%x , cur_level= %u , last_level= %u",
				addr,current->ktau->cur_level, current->ktau->last_level);
#endif /*CONFIG_KTAU_MERGE*/

	if(current->ktau->last_level == 0 && 
			current->ktau->cur_level == 0){
		/* Normal Case */
		current->ktau->cur_level = 1;
#ifdef CONFIG_KTAU_MERGE
		if(current->ktau->ppstate)info("---- Normal case");
#endif /*CONFIG_KTAU_MERGE*/
	}else if(current->ktau->last_level < current->ktau->cur_level){
		/* Enter Nested Case */
		current->ktau->last_level++;
		current->ktau->cur_level++;
#ifdef CONFIG_KTAU_MERGE
		if(current->ktau->ppstate)info("---- Enter nested case");
#endif /*CONFIG_KTAU_MERGE*/
	}else if(current->ktau->last_level > current->ktau->cur_level){
		/* Re-enter Nested Case */
		current->ktau->last_level--;
		current->ktau->cur_level++;
#ifdef CONFIG_KTAU_MERGE
		if(current->ktau->ppstate)info("---- Re-enter nested case");
#endif /*CONFIG_KTAU_MERGE*/
	}else{
		info("ktau_start_prof: Invalid instrumentation level");	
	}

	if(!(cur_ent)){
		current->ktau->ktau_hash[index] = kmalloc(sizeof(h_ent),GFP_ATOMIC);
		current->ktau->ent_count++;
		cur_ent = current->ktau->ktau_hash[index];
		cur_ent->type = KTAU_TIMER;
		cur_ent->data.timer.count 	= 0;
		cur_ent->data.timer.incl 	= 0;
		cur_ent->data.timer.excl 	= 0;
		cur_ent->data.timer.addr  	= addr;
	}
	cur_ent->data.timer.start = ktau_rdtsc();

#ifdef CONFIG_KTAU_IRQ
	ktau_irq_enable_restore_flags_unlock(flags, &current->ktau->hash_lock);
#else /* CONFIG_KTAU_IRQ */
	spin_unlock(&current->ktau->hash_lock);
	ktau_bh_enable();
#endif /* CONFIG_KTAU_IRQ */
}

/*
 * This fuction is the stop profile
 */
void ktau_stop_prof(unsigned int index, unsigned int addr){
	unsigned long flags = 0;
	unsigned long long diff_time = 0;

#ifdef CONFIG_KTAU_IRQ
	ktau_irq_disable_save_flags_lock(flags, &current->ktau->hash_lock);
#else /* CONFIG_KTAU_IRQ */
	local_bh_disable();
	spin_lock(&current->ktau->hash_lock);
#endif /* CONFIG_KTAU_IRQ */

	h_ent *cur_ent = (h_ent*)(current->ktau->ktau_hash[index]);
	if((cur_ent)){
		cur_ent->data.timer.count++;
		diff_time = ktau_rdtsc() - cur_ent->data.timer.start;
		cur_ent->data.timer.incl += diff_time;
		cur_ent->data.timer.start = 0;

#ifdef CONFIG_KTAU_MERGE
		if(current->ktau->ppstate) info("STOPPING: addr=%x , cur_level= %u , last_level= %u",
					addr,current->ktau->cur_level, current->ktau->last_level);
#endif /*CONFIG_KTAU_MERGE*/

		/* Incl / Excl time logic */	
		if(current->ktau->last_level == 0 && 
				current->ktau->cur_level == 1){
			/*Normal case */
			current->ktau->last_level 	= 0;
			current->ktau->cur_level 	= 0;
			cur_ent->data.timer.excl  	+= diff_time;

#ifdef CONFIG_KTAU_MERGE
			if(current->ktau->ppstate) info("---- Normal Case");
#endif /*CONFIG_KTAU_MERGE*/
		}else if(current->ktau->last_level == 2 && 
				current->ktau->cur_level == 1){
			/* Leaving to level 0 */
			cur_ent->data.timer.excl 	+= 
				diff_time - current->ktau->accum[current->ktau->last_level];
			current->ktau->accum[current->ktau->last_level] = 0;
			current->ktau->accum[current->ktau->cur_level]  = 0;
			current->ktau->last_level 	= 0;
			current->ktau->cur_level 	= 0;

#ifdef CONFIG_KTAU_MERGE
			if(current->ktau->ppstate) info("---- Leaving to level 0");
#endif /*CONFIG_KTAU_MERGE*/
		}else if(current->ktau->last_level < current->ktau->cur_level){
			/* Leaving from inner most nested */
			cur_ent->data.timer.excl 	+= diff_time;
			current->ktau->accum[current->ktau->cur_level]  = diff_time;
			current->ktau->last_level++;
			current->ktau->cur_level--;
	
#ifdef CONFIG_KTAU_MERGE
			if(current->ktau->ppstate) info("---- Leaving from inner most nested");
#endif /*CONFIG_KTAU_MERGE*/
		}else if(current->ktau->last_level > current->ktau->cur_level){
			/* Leaving Nested */
			cur_ent->data.timer.excl 	+= 
				diff_time - current->ktau->accum[current->ktau->last_level];
			current->ktau->accum[current->ktau->last_level] = 0;
			current->ktau->accum[current->ktau->cur_level]  += diff_time;
			current->ktau->last_level--;
			current->ktau->cur_level--;

#ifdef CONFIG_KTAU_MERGE
			if(current->ktau->ppstate)info("---- Leaving Nested");
#endif /*CONFIG_KTAU_MERGE*/
		}else{
			info("ktau_stop_prof: Invalid instrumentation level");	
		}
	}

#ifdef CONFIG_KTAU_IRQ
	ktau_irq_enable_restore_flags_unlock(flags, &current->ktau->hash_lock);
#else /* CONFIG_KTAU_IRQ */
	spin_unlock(&current->ktau->hash_lock);
	ktau_bh_enable();
#endif /* CONFIG_KTAU_IRQ */

#ifdef CONFIG_KTAU_MERGE
	/* Update ktau_state once get to the outer most level */
	if(current->ktau->cur_level == 0 && current->ktau->last_level == 0)
		if(current->ktau->ppstate) update_userstate(diff_time, current->ktau->ppstate);
#endif /*CONFIG_KTAU_MERGE*/
}

/*
 * This fuction is the event profile
 */
void ktau_event_prof(unsigned int index, unsigned int addr){
	unsigned long flags = 0;
	h_ent *cur_ent = (h_ent*)(current->ktau->ktau_hash[index]);

#ifdef CONFIG_KTAU_IRQ
	ktau_irq_disable_save_flags_lock(flags, &current->ktau->hash_lock);
#else /* CONFIG_KTAU_IRQ */
	local_bh_disable();
	spin_lock(&current->ktau->hash_lock);
#endif /* CONFIG_KTAU_IRQ */

	if(!(cur_ent)){
		current->ktau->ktau_hash[index] = kmalloc(sizeof(h_ent),GFP_ATOMIC);
		current->ktau->ent_count++;
		cur_ent = current->ktau->ktau_hash[index];
		cur_ent->type = KTAU_TIMER;
		cur_ent->data.timer.count = 0;
		cur_ent->data.timer.incl = 0;
		cur_ent->data.timer.start = 0;
		cur_ent->data.timer.addr  = addr;
	}
	cur_ent->data.timer.count++;

#ifdef CONFIG_KTAU_IRQ
	ktau_irq_enable_restore_flags_unlock(flags, &current->ktau->hash_lock);
#else /* CONFIG_KTAU_IRQ */
	spin_unlock(&current->ktau->hash_lock);
	ktau_bh_enable();
#endif /* CONFIG_KTAU_IRQ */
}

