/*************************************************************
 * File         : user-src/src/runktau/runktau.c
 * Version      : $Id: runktau.c,v 1.1 2005/09/28 19:47:39 anataraj Exp $
 *
 ************************************************************/

/************************************************
 * KTAU Performance Analysis Suite		*
 * 						*
 * runktau:					*
 * 	- Launches any program,			*
 * 	- waits for termination,		*
 * 	- reads Kernel Profile/Trace,		*
 * 	- and dumps to file.			*
 *						*
 * 	- Functionality similar to 'time' cmd	*
 * 						*
 * Usage: runktau <full program path> [args] [args]
 * 						*
 ***********************************************/

#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>

#include <ktau_proc_interface.h>

/* Status Field */
#define S_INPROG	0x0
#define S_DONE		0x1

#define POLL_INTV	5	/*Poll every 5 secs for result */

#define FNAME_SIZE 256

volatile int status = S_INPROG;

volatile int numchildren = 0;

void dump_ktau_data(int pid, int type); 

static void _sigchld(int readpid)
{
	int pid;
	int saved_errno = errno;

	//printf("Calling waitpid with: %d\n", readpid);

	while ((pid = waitpid(readpid, 0, WNOHANG)) != 0) {

		//printf("Waitpid returned: %d\n",pid);

		if (pid == -1) {
			if (errno != EINTR)
				break;
		} else
			++numchildren;
	}

	errno = saved_errno;
	status = S_DONE;
}

static void sigchld() {
	_sigchld(-1);
}

static void sigchld_sigaction(int signo, siginfo_t *psiginfo, void* parg) {
	int pid = 0;

	if(signo != SIGCHLD) {
		printf("Non-SigChld Signal: %d. Ignoring.\n", signo);
		return;
	}

	//printf("SigChld Recieved....\n");

	if(!psiginfo) {
		printf("psigninfo is NULL. Ignoring.\n");
		return;
	}

	pid = psiginfo->si_pid;

	//printf("Pid form sigingfo: %d\n", pid);
	
	dump_ktau_data(pid, KTAU_TYPE_PROFILE);

	_sigchld(pid);
}



int main(int argc, char* argv[])
{
	struct sigaction act;
	int pid;
	char *parent="Parent:", *child="Child:";
	char* name = parent;

	act.sa_sigaction = sigchld_sigaction;
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_SIGINFO;
	sigaction(SIGCHLD, &act, 0);

	switch((pid=fork())) {
		case -1:
			perror("fork");
			return 1;
		case 0:
			name = child;
			printf("%s] Error: Execv returned: %d\n", name, execv(argv[1], argv+1));
			_exit(0);
			break;
		default:
			//printf("%s Parent: Child Pid is: %d\n", name, pid);
			while(status == S_INPROG)
				sleep(POLL_INTV);
			break;
	}

	//printf("%s Number of children = %d\n", name, numchildren);
	return 0;
}


void dump_ktau_data(int pid, int type) {
	char *buffer, fname[FNAME_SIZE];
        ktau_output* output = NULL;
        int size = 0, ret_val = 0, rd_size = 0, noprofs = 0;
	FILE* dump_fs;

	size = read_size(type, 0, &pid, 1, 0, NULL, -1.0);

	if(size <= 0) {
		printf("dump_ktau_data: read_size ret bad size: %d.\n", size);
		return;
	}

	buffer = (char*)malloc(size);
	if(!buffer) {
		printf("dump_ktau_data: malloc failed. \n");
		return;
	}


        /* Read Data*/
        ret_val = read_data(type, 0, &pid, 1, buffer, size,
                        0 /*tolerance*/,
                        NULL /*flags*/);
        if(ret_val < 0) {
                perror("dump_ktau_data: read_data failure.");
		goto free_out;
        }

        /* Unpack Data*/
        rd_size = ret_val;
        noprofs = unpack_bindata(type, buffer, rd_size, &output);
        if(!noprofs || !output) {
                printf("ktau_dump_data: ktau_output from unpack_bindata is:%x. no-profs is:%d\n", output, noprofs);
		goto free_out;
        }
	
	sprintf(fname, "./Kprofile.%d", pid);
	dump_fs = fopen(fname, "w");
	if(!dump_fs) {
		printf("dump_ktau_data: Cannot open %s. fopen returned Null.", fname);
		goto free_out;
	}

	print_ktau_output(type, dump_fs, output);

free_out:
	free(buffer);
	buffer = NULL;

        return;
}

