/****************************************************************************
**			TAU Portable Profiling Package			   **
**			http://www.acl.lanl.gov/tau		           **
*****************************************************************************
**    Copyright 1997  						   	   **
**    Department of Computer and Information Science, University of Oregon **
**    Advanced Computing Laboratory, Los Alamos National Laboratory        **
****************************************************************************/
/***************************************************************************
**	File 		: TauKtau.h					  **
**	Description 	: TAU Kernel Profiling Interface		  **
**	Author		: Suravee Suthikulpanit				  **
**	Contact		: suravee@cs.uoregon.edu                    	  **
**	Flags		: Compile with				          **
**			  -DTAU_KTAU to enable KTAU	                  **
**	Documentation	: 					          **
***************************************************************************/

#ifndef _TAUKTAU_H_
#define _TAUKTAU_H_

#define NAME_SIZE	50
#define MAP_SIZE	10 * 1020

#include "Profile/ktau_proc_interface.h"
/////////////////////////////////////////////////////////////////////
//
// class TauKtau
//
//////////////////////////////////////////////////////////////////////

typedef struct _addr_map{
	unsigned long long addr;
	char name[NAME_SIZE];
}addr_map;


class TauKtau 
{
public:
	char *startBuf;
	char *stopBuf;
	unsigned int startSize;
	unsigned int stopSize;
	unsigned int outSize;
	unsigned long kallsyms_size;
	addr_map kallsyms_map[MAP_SIZE];
	
	// Constructor
	TauKtau();

	// Destructor
	~TauKtau();
	
	// APIs
	int TauKtau::StartKProfile();
	int TauKtau::StopKProfile();
	int TauKtau::DumpKProfile(ktau_output **diffOutput);
private:
	char* TauKtau::MapKallsyms(o_ent *out, unsigned *cur_index);
	int TauKtau::ReadKallsyms();
	int TauKtau::DiffKProfile(ktau_output **diffOutput);
};


#endif /* _TAUKTAU_H_*/
/***************************************************************************
 * $RCSfile: FunctionInfo.h,v $   $Author: sameer $
 * $Revision: 1.30 $   $Date: 2005/01/05 22:57:10 $
 * POOMA_VERSION_ID: $Id: FunctionInfo.h,v 1.30 2005/01/05 22:57:10 sameer Exp $ 
 ***************************************************************************/
