/****************************************************************************
**			TAU Portable Profiling Package			   **
**			http://www.acl.lanl.gov/tau		           **
*****************************************************************************
**    Copyright 1999  						   	   **
**    Department of Computer and Information Science, University of Oregon **
**    Advanced Computing Laboratory, Los Alamos National Laboratory        **
****************************************************************************/
/***************************************************************************
**	File 		: TauKtau.cpp					  **
**	Description 	: TAU Kernel Profiling 				  **
**	Author		: Surave Suthikulpanit				  **
**	Contact		: suravee@cs.uoregon.edu		 	  **
**	Flags		: Compile with				          **
**			  -DTAUKTAU 					  **
**	Documentation	:                                                 **
***************************************************************************/

#include <iostream.h>
#include <fstream.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "Profile/TauKtau.h"
#include "Profile/ktau_proc_interface.h"

#define LINE_SIZE		1024	
#define KALLSYMS_PATH		"/proc/kallsyms"
#define OUTPUT_NAME_SIZE	100

/*-------------------------- CON/DESTRUCTOR ---------------------------*/
TauKtau::TauKtau()
{
        startBuf = NULL;
        stopBuf  = NULL;
        startSize= 0;
        stopSize = 0;
        kallsyms_size = 0;
}

TauKtau::~TauKtau()
{

}

/*----------------------------- PUBLIC --------------------------------*/
int TauKtau::StartKProfile()
{
	startSize = read_size(NULL, 0, 0, NULL, -1);
	startBuf = (char*)malloc(startSize*sizeof(ktau_output));
	startSize = read_data(NULL, 0, startBuf, startSize, 0, NULL);
	return startSize;
}

int TauKtau::StopKProfile()
{
	unsigned int ret = 0;
	ktau_output *diffOutput;
	stopSize = read_size(NULL, 0, 0, NULL, -1);
	stopBuf = (char*)malloc(stopSize*sizeof(ktau_output));
	stopSize = read_data(NULL, 0, stopBuf, stopSize, 0, NULL);

	if((outSize = DiffKProfile(&diffOutput)) < 0){
		return(-1);	
	}
	if((ret = DumpKProfile(&diffOutput)) < 0){
		return(-1);	
	};
	return(ret);
}

char* TauKtau::MapKallsyms(o_ent *out,unsigned int *cur_index){
	int i=0;
	for(i=0;i<kallsyms_size;i++){
		if(out->entry.data.timer.addr == kallsyms_map[i].addr){
			return(kallsyms_map[i].name);	
		}
	}
	return("");
	
}

int TauKtau::DumpKProfile(ktau_output **diffOutput)
{
	int i=0,j=0;
	char output_path[OUTPUT_NAME_SIZE];
	o_ent *ptr;
	unsigned int cur_index = 0;

	ReadKallsyms();
       
	/* 
	 * Create output directory ./Kprofile 
	 */
	sprintf(output_path,"./Kprofile");
	if(mkdir(output_path,755) == -1){
		perror("mkdir");
		return(-1);
	}

	/*
	 * Dumping output to the output file "./Kprofile/profile.<pid>.0.0"
	 */
	// Data format :
        // %d templated_functions
        // "%s %s" %ld %G %G  
        //  funcname type numcalls Excl Incl
        // %d aggregates
        // <aggregate info>
#define mapKey(x)	((x - kallsyms_map[0].addr)>>2)

	for(i=0;i<outSize;i++){
		printf("DumpKProfile: Dumping pid = %d size = %d\n",
				(*diffOutput+i)->pid,
				(*diffOutput+i)->size);
		sprintf(output_path,"./Kprofile/profile.%u.0.0",(*diffOutput+i)->pid);
		ofstream fs_output (output_path , ios::out);
		if(!fs_output.is_open()){
			cout << "Error opening file: " << output_path << "\n";
			return(-1);
		}
		
		fs_output << ((*diffOutput)+i)->size << " templated_functions" << endl;
		fs_output << "# Name Calls Subrs Excl Incl ProfileCalls" << endl;

		for(j=0;j < (*diffOutput+i)->size; j++){
			ptr = (((*diffOutput)+i)->ent_lst)+j;
			fs_output << "\"" << 
				MapKallsyms(ptr,&cur_index) << "()\" " 		//Name
				  << ptr->entry.data.timer.count << " "		//Calls
				  << 0 << " "					//Subrs
				  << ptr->entry.data.timer.total << " "		//Excl
				  << ptr->entry.data.timer.total << " "		//Incl
				  << 0 << " "					//ProfileCalls
				  << "GROUP=\"TAU_DEFAULT\""
				  << endl;							
		}
			
		fs_output << 0 << " aggregates" << endl;
		fs_output.close();
	}
	return 0;
}
/*----------------------------- PRIVATE -------------------------------*/
int diff_h_ent(o_ent *o1, o_ent *o2){
	printf("diff_h_ent: o1 -> count = %u total = %llu\n", 
			o1->entry.data.timer.count,
			o1->entry.data.timer.total);
	printf("diff_h_ent: o2 -> count = %u total = %llu\n", 
			o2->entry.data.timer.count,
			o2->entry.data.timer.total);

	/*
	 * Comparing h_ent of KTAU_TIMER type.
	 */
	if(o1->entry.type == o2->entry.type && o1->entry.type == KTAU_TIMER){
		if(o1->entry.data.timer.addr == o2->entry.data.timer.addr){
			o2->entry.data.timer.count -= o1->entry.data.timer.count;
			o2->entry.data.timer.total -= o1->entry.data.timer.total;
			printf("diff_h_ent: diff -> count = %u total = %llu\n", 
				o2->entry.data.timer.count,
				o2->entry.data.timer.total);
			return(0);
		}
	}
	return(-1);
}

int diff_ktau_output(ktau_output *start, ktau_output *stop, ktau_output *out){
	int i,j,k;
	o_ent *st;
	o_ent *sp;

	out->ent_lst = (o_ent*)malloc(sizeof(o_ent)*stop->size);
	/*
	 * For each o_ent, we have to check the hash-index. If the index
	 * doesn't exist in out, then just diff with 0.
	 * */
	for(i=0,j=0,k=0; i<start->size || j<stop->size;){
		st=(start->ent_lst+i);
		sp=(stop->ent_lst+j);
		if(st->index == sp->index){
			if(diff_h_ent(st,sp)){
				return(-1);
			}
			memcpy(((out->ent_lst)+k), ((stop->ent_lst)+j), sizeof(o_ent));
			i++;j++;k++;
		}else{
			memcpy(((out->ent_lst)+k), ((stop->ent_lst)+j), sizeof(o_ent));
			j++;k++;
		}
	}
	return(k);
}

#define maxof(x,y)	((x>=y)?x:y)
int TauKtau::DiffKProfile(ktau_output **diffOutput)
{
	int i,j,k,diff_count;
	ktau_output *startOutput;
	ktau_output *stopOutput;
	unsigned int startProcNum = unpack_bindata(startBuf, startSize, &startOutput);	
	unsigned int stopProcNum  = unpack_bindata(stopBuf, stopSize, &stopOutput);	
	printf("DiffKProfile: startProcNum = %u\n",startProcNum);
	printf("DiffKProfile: stopProcNum = %u\n",stopProcNum);
	*diffOutput = (ktau_output*)malloc(sizeof(ktau_output)*maxof(startProcNum,stopProcNum));
	
	/*
	 * For each ktau_output, we have to check pid if it is the same.
	 * Otherwise, we just ignore that one.
	 */
	for(i=0,j=0,k=0; i<startProcNum & j<stopProcNum;){
		if((startOutput+i)->pid == (stopOutput+j)->pid){
			printf("DiffKProfile: Diffing pid %d\n",(startOutput+i)->pid);
			printf("DiffKProfile: start size = %d\n",(startOutput+i)->size);
			printf("DiffKProfile: stop size  = %d\n",(stopOutput+i)->size);
			diff_count = diff_ktau_output((startOutput+i),(stopOutput+j),((*diffOutput)+k));
			(*diffOutput+k)->size = diff_count;
			(*diffOutput+k)->pid = (startOutput+i)->pid;
			i++;j++;k++;
		}else if((startOutput+i)->pid < (stopOutput+j)->pid){
			printf("DiffKProfile: pid unmatch. Increment start.\n");
			i++;
		}else{
			printf("DiffKProfile: pid unmatch. Increment stop.\n");
			j++;
		}
	}
	return(k);
}

int TauKtau::ReadKallsyms()
{
	char line[LINE_SIZE];
	char *ptr = line;
	char *addr;
	char *flag;
	char *func_name;
	unsigned int stext = 0;	
	int found_stext = 0;	
		
	/* Open and read file */
	ifstream fs_kallsyms (KALLSYMS_PATH,ios::in);
	if(!fs_kallsyms.is_open()){
		cout << "Error opening file: " << KALLSYMS_PATH << "\n";
		return(-1);
	}

	/* Tokenize each line*/
	while(!fs_kallsyms.eof()){
		fs_kallsyms.getline(line,LINE_SIZE);
		ptr = line;
		addr = strtok(ptr," ");
		flag = strtok(NULL," ");
		func_name = strtok(NULL," ");
		
		if(!strcmp("T",flag) || !strcmp("t",flag)){
			if(!found_stext && !strcmp("stext",func_name)){
				kallsyms_map[kallsyms_size].addr = strtoll(addr,NULL,16);
				strcpy(kallsyms_map[kallsyms_size++].name, func_name);
				found_stext = 1;
				//printf("ReadKallsyms: address %llx , name %s \n",
				//		kallsyms_map[kallsyms_size-1].addr,
				//		kallsyms_map[kallsyms_size-1].name);
			} else{
				kallsyms_map[kallsyms_size].addr = strtoll(addr,NULL,16);
				strcpy(kallsyms_map[kallsyms_size++].name, func_name);
				//printf("ReadKallsyms: address %llx , name %s \n",
				//		kallsyms_map[kallsyms_size-1].addr,
				//		kallsyms_map[kallsyms_size-1].name);
			}	
		}else if(!strcmp("_etext",func_name)){
			kallsyms_map[kallsyms_size].addr = strtoll(addr,NULL,16);
			strcpy(kallsyms_map[kallsyms_size++].name, func_name);
			//printf("ReadKallsyms: address %llx , name %s \n",
				//		kallsyms_map[kallsyms_size-1].addr,
				//		kallsyms_map[kallsyms_size-1].name);
			break;	
		}	
	}
	/* Close /proc/kallsyms */	
	fs_kallsyms.close();

	return kallsyms_size;
}

/***************************************************************************
 * $RCSfile: TauKtau.cpp,v $   $Author: suravee $
 * $Revision$   $Date$
 ***************************************************************************/

	





